class TermColor:
    ERROR_HIGHLIGHT = "\u001b[41m"  # red background
    OK_HIGHLIGHT = "\u001b[42m"  # green background
    WARNING_HIGHLIGHT = "\u001b[43m"  # yellow background
    ERROR = "\033[38;5;160m"  # red
    OK = "\033[38;5;46m"  # green
    WARNING = "\033[38;5;220m"
    ENDC = "\033[0m"



import ast

import astor
import black


class Walker(ast.NodeTransformer):
    __imports = {}
    __sv = set()  # serializers and views

    def __init__(self, file, options=None):
        self.file = file
        with open(self.file, "w+") as f:
            self.tree = ast.parse(f.read())
        self.options = options
        super().__init__()

    def visit_Attribute(self, node):
        ast.NodeVisitor.generic_visit(self, node)
        if isinstance(node.value, ast.Name) and node.value.id == "models":
            self.__sv.add(node.attr)
        return node

    def visit_Assign(self, node):
        ast.NodeVisitor.generic_visit(self, node)
        if self.options:
            if node.targets[0].id == self.options["variable"]:
                node.value.elts.extend(
                    [
                        ast.Constant(value=x, kind=None)
                        for x in self.options["variable_values"]
                        if x not in [app.value for app in node.value.elts]
                    ]
                )
        return node

    def visit_Import(self, node):
        ast.NodeVisitor.generic_visit(self, node)
        self.__imports["free"] = [x.name for x in node.names]
        return node

    def visit_ImportFrom(self, node):
        ast.NodeVisitor.generic_visit(self, node)
        self.__imports[node.module] = [x.name for x in node.names]
        return node

    def get_imports(self):
        self.visit(self.tree)
        return self.__imports

    def get_sv(self):
        self.visit(self.tree)
        return self.__sv

    def mutate(self):
        with open(self.file, "w") as f:
            content = astor.to_source(self.visit(self.tree))
            f.write(
                black.format_file_contents(content, fast=False, mode=black.FileMode())
            )



from enum import Enum

from django.template import Context, Template

from scaffold.kit.patterns import (
    CHAR_FIELD_TEMPLATE,
    DECIMAL_FIELD_TEMPLATE,
    FIELD_TEMPLATE,
    FOREIGN_KEY_TEMPLATE,
    MANY_TO_MANY_FIELD_TEMPLATE,
    MODEL_TEMPLATE,
    ONE_TO_ONE_FIELD_TEMPLATE,
    SERIALIZER_TEMPLATE,
    VIEW_SET_URL_TEMPLATE,
    VIEW_SET_VIEW_TEMPLATE,
)


class Field(Enum):
    Genl = (FIELD_TEMPLATE, ["name", "type"])
    Char = (CHAR_FIELD_TEMPLATE, ["name", "type", "max"])
    Foreign = (FOREIGN_KEY_TEMPLATE, ["name", "type", "model", "delete"])
    ManyToMany = (MANY_TO_MANY_FIELD_TEMPLATE, ["name", "type", "model", "delete"])
    OneToOne = (ONE_TO_ONE_FIELD_TEMPLATE, ["name", "type", "model", "delete"])
    Decimal = (DECIMAL_FIELD_TEMPLATE, ["name", "type", "max", "places"])


class FieldTemplate:
    @staticmethod
    def convert(args):
        type = args[1]
        try:
            pattern, options = Field[type].value[0], Field[type].value[1]
            context = dict(zip(options, args))
        except KeyError:
            pattern, options = Field.Genl.value[0], Field.Genl.value[1]
            context = dict(zip(options, args))
        return Template(pattern).render(context=Context(context))


class ModelTemplate:
    @staticmethod
    def convert(context):
        return Template(MODEL_TEMPLATE).render(context=Context(context))


class SerializerTemplate:
    @staticmethod
    def convert(context):
        return Template(SERIALIZER_TEMPLATE).render(context=Context(context))


class UrlTemplate:
    @staticmethod
    def convert(context):
        return Template(VIEW_SET_URL_TEMPLATE).render(context=Context(context))


class ViewTemplate:
    @staticmethod
    def convert(context):
        return Template(VIEW_SET_VIEW_TEMPLATE).render(context=Context(context))



import subprocess
import sys

from django.core.management import CommandError

from scaffold.kit.colors import TermColor
from scaffold.kit.templates import (
    FieldTemplate,
    ModelTemplate,
    SerializerTemplate,
    UrlTemplate,
    ViewTemplate,
)
from scaffold.kit.utils import Walker


class Scaffold:
    def __init__(
        self,
        proj_settings,
        app_config,
        new_model,
        fields,
        serializers,
        urls,
        views,
    ):
        self.proj_settings = proj_settings
        self.new_model = new_model
        self.app_config = app_config
        self.models = self.get_model_names()
        self.fields = fields
        self.serializers = serializers
        self.urls = urls
        self.views = views

    def get_model_names(self):
        return [m.__name__ for m in self.app_config.get_models()]

    def get_field(self, field):
        args = field.split(":")
        return FieldTemplate.convert(args)

    def check_models(self, models):
        missing_models = [x for x in models if x not in set(self.get_model_names())]
        return missing_models

    def check_sv(self, file, sv):
        existing_sv = Walker(file).get_sv()
        excess_sv = [x for x in sv if x in existing_sv]
        return excess_sv

    def create_model(self):
        if self.new_model in self.get_model_names():
            raise CommandError(f"model {self.new_model} already exists...")
        fields = []
        for field in self.fields:
            new_field = self.get_field(field)
            fields.append(new_field)
        with open(self.app_config.models_module.__file__, "a") as mf:
            content = ModelTemplate.convert(
                context={"name": self.new_model, "fields": fields}
            )
            mf.write(content)
        subprocess.call(["black", self.app_config.models_module.__file__, "-q"])
        print(f"{TermColor.OK}model: {self.new_model} has been created{TermColor.ENDC}")

    def check_imports(self, filename, imports):
        existing_imports = Walker(file=filename).get_imports()
        missing_imports = {}
        for key, value in imports.items():
            missing_values = [
                x for x in value if x not in set(existing_imports.get(key, []))
            ]
            if missing_values:
                missing_imports[key] = missing_values
        return missing_imports

    def create_serializers(self):
        serializer_file_path = f"{self.app_config.module.__path__[0]}/serializers.py"
        serializers = (
            self.get_model_names() if not self.serializers else self.serializers
        )

        missing_models = self.check_models(serializers)
        if missing_models:
            error = (
                f'{" ".join(missing_models)} do not exist...'
                if len(missing_models) > 1
                else f'{" ".join(missing_models)} does not exist...'
            )
            raise CommandError(error)

        excess_serializers = self.check_sv(serializer_file_path, serializers)
        if excess_serializers:
            error = (
                f'{" ".join(excess_serializers)} already exist...'
                if len(excess_serializers) > 1
                else f'{" ".join(excess_serializers)} already exists...'
            )
            raise CommandError(error)

        missing_imports = self.check_imports(
            serializer_file_path,
            {"rest_framework": ["serializers"], f"{self.app_config.name}": ["models"]},
        )
        with open(serializer_file_path, "a") as sf:
            content = SerializerTemplate.convert(
                context={"models": serializers, "imports": missing_imports}
            )
            sf.write(content)
        subprocess.call(["black", serializer_file_path, "-q"])
        print(
            f"{TermColor.OK}serializers: {' '.join(serializers)} have been created{TermColor.ENDC}"
        ) if len(serializers) > 1 else print(
            f"{TermColor.OK}serializer: {' '.join(serializers)} has been created{TermColor.ENDC}"
        )

    def create_urls(self):
        url_file_path = f"{self.app_config.module.__path__[0]}/urls.py"
        existing_models = self.get_model_names()
        with open(url_file_path, "w+") as uf:
            content = UrlTemplate.convert(
                context={"app": self.app_config.name, "models": existing_models}
            )
            uf.write(content)
        subprocess.call(["black", url_file_path, "-q"])
        print(
            f"{TermColor.OK}urls: SimpleRouter for all models has been created{TermColor.ENDC}"
        )

    def create_views(self):
        view_file_path = f"{self.app_config.module.__path__[0]}/views.py"
        views = self.get_model_names() if not self.views else self.views

        missing_models = self.check_models(views)
        if missing_models:
            raise CommandError(f'{" ".join(missing_models)} do/does not exist...')

        excess_views = self.check_sv(view_file_path, views)
        if excess_views:
            raise CommandError(f'{" ".join(excess_views)} do/does exist...')

        missing_imports = self.check_imports(
            view_file_path,
            {
                "django.shortcuts": ["get_object_or_404"],
                "rest_framework": ["viewsets", "response"],
                f"{self.app_config.name}": ["models", "serializers"],
            },
        )
        with open(view_file_path, "a") as wf:
            content = ViewTemplate.convert(
                context={"models": views, "imports": missing_imports}
            )
            wf.write(content)
        subprocess.call(["black", view_file_path, "-q"])
        print(
            f"{TermColor.OK}view: {' '.join(views)} have been created{TermColor.ENDC}"
        ) if len(views) > 1 else print(
            f"{TermColor.OK}view: {' '.join(views)} has been created{TermColor.ENDC}"
        )

    def execute(self):
        if self.new_model:
            self.create_model()
        if self.urls:
            self.create_urls()
        if self.serializers is not None:
            self.create_serializers()
        if self.views is not None:
            self.create_views()


class ScaffoldApp:
    def __init__(self, proj_settings, new_apps):
        self.apps = new_apps
        self.proj_settings = proj_settings

    def create_app(self):
        for app in self.apps:
            try:
                subprocess.call(["python", "manage.py", "startapp", app])
            except Exception as e:
                print(e)
        walker = Walker(
            file=sys.modules[self.proj_settings].__file__,
            options={"variable": "INSTALLED_APPS", "variable_values": self.apps},
        )
        walker.mutate()

    def execute(self):
        if self.apps:
            self.create_app()



FIELD_TEMPLATE = "{{ name }} = models.{{ type|title }}Field()"

DECIMAL_FIELD_TEMPLATE = (
    "{{ name }} = models.{{ type }}Field(max_digits={{ max|default:'5' }}, decimal_places={{ "
    "places|default:'2' }}) "
)

CHAR_FIELD_TEMPLATE = (
    "{{ name }} = models.{{ type }}Field(max_length={{max|default:'255'}})"
)

FOREIGN_KEY_TEMPLATE = (
    "{{ name }} = models.{{ type }}Key({{ model|default:\"'self'\" }}, on_delete=models.{{ "
    'delete|default:"CASCADE" }}) '
)

ONE_TO_ONE_FIELD_TEMPLATE = (
    "{{ name }} = models.{{ type }}Field({{ model|default:\"'self'\" }}, on_delete=models.{{ "
    'delete|default:"CASCADE" }}) '
)

MANY_TO_MANY_FIELD_TEMPLATE = (
    "{{ name }} = models.{{ type }}Field({{ model|default:\"'self'\" }})"
)

MODEL_TEMPLATE = """
class {{ name }}(models.Model):{% for field in fields %}
    {{field}} {% endfor %}
    update_date = models.DateTimeField(auto_now=True)
    create_date = models.DateTimeField(auto_now_add=True)
    class Meta:
        ordering = ['-id']
"""

SERIALIZER_TEMPLATE = """{% for key, value  in imports.items %}from {{ key }} import {{ value|join:", " }} 
{% endfor %}
{% for model in models %}
class {{ model }}Serializer(serializers.ModelSerializer):
    class Meta:
        model = models.{{ model }}
        fields = '__all__'
        
{% endfor %}
"""

VIEW_SET_URL_TEMPLATE = """from rest_framework.routers import SimpleRouter
from {{ app }} import views


router = SimpleRouter()
{% for model in models %}
router.register(r'{{ model | lower }}', views.{{ model }}ViewSet, '{{model}}'){% endfor %}
urlpatterns = router.urls
"""

VIEW_SET_VIEW_TEMPLATE = """{% for key, value  in imports.items %}from {{ key }} import {{ value|join:", " }} 
{% endfor %}
{% for model in models %}
class {{ model }}ViewSet(viewsets.ViewSet):
    def list(self, request):
        queryset = models.{{ model }}.objects.order_by('pk')
        serializer = serializers.{{ model }}Serializer(queryset, many=True)
        return response.Response(serializer.data)
    
    def create(self, request):
        serializer = serializers.{{ model }}Serializer(data=request.data)
        if serializer.is_valid():
            serializer.save()
            return response.Response(serializer.data, status=201)
        return response.Response(serializer.errors, status=400)
    
    def retrieve(self, request, pk=None):
        queryset = models.{{ model }}.objects.all()
        item = get_object_or_404(queryset, pk=pk)
        serializer = serializers.{{ model }}Serializer(item)
        return response.Response(serializer.data)
    
    def update(self, request, pk=None):
        try:
            item = models.{{ model }}.objects.get(pk=pk)
        except models.{{ model }}.DoesNotExist:
            return response.Response(status=404)
        serializer = serializers.{{ model }}Serializer(item, data=request.data)
        if serializer.is_valid():
            serializer.save()
            return response.Response(serializer.data)
        return response.Response(serializer.errors, status=400)
    
    def destroy(self, request, pk=None):
        try:
            item = models.{{ model }}.objects.get(pk=pk)
        except models.{{ model }}.DoesNotExist:
            return response.Response(status=404)
        item.delete()
        return response.Response(status=204)
        
{% endfor %}
"""



import os

from django.core.management.base import AppCommand

from scaffold.scaffold import Scaffold


class Command(AppCommand):
    def get_version(self):
        return f"scaffold 0.1.3"

    def create_parser(self, prog_name, subcommand, **kwargs):
        parser = super(Command, self).create_parser(prog_name, subcommand)
        return parser

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument(
            "-m",
            "--model",
            dest="model",
            default=None,
            nargs="+",
            help="Add a new model with specific fields",
        )
        parser.add_argument(
            "-s",
            "--serializers",
            dest="serializers",
            default=None,
            nargs="*",
            help="Add a new serializer for the specific model or use keyword 'a' for all models",
        )
        parser.add_argument(
            "-u",
            "--urls",
            dest="urls",
            action="store_true",
            help="Add urls for all models",
        )
        parser.add_argument(
            "-vi",
            "--views",
            dest="views",
            default=None,
            nargs="*",
            help="Add a view for the specific model or use keyword 'a' for all models",
        )

    def handle_app_config(self, app_config, **options):
        settings = (
            options["settings"]
            if options.get("settings")
            else os.environ.get("DJANGO_SETTINGS_MODULE")
        )
        new_model = options["model"][0] if options.get("model") else None
        fields = options["model"][1:] if options.get("model") else None
        serializers = options.get("serializers", None)
        views = options.get("views")
        urls = options.get("urls", False)

        scaffold = Scaffold(
            proj_settings=settings,
            app_config=app_config,
            new_model=new_model,
            fields=fields,
            views=views,
            serializers=serializers,
            urls=urls,
        )
        scaffold.execute()



import os

from django.core.management.base import BaseCommand

from scaffold.scaffold import ScaffoldApp


class Command(BaseCommand):
    def get_version(self):
        return f"scaffold 0.1.3"

    def create_parser(self, prog_name, subcommand, **kwargs):
        parser = super(Command, self).create_parser(prog_name, subcommand)
        return parser

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument("new_apps", nargs="*", help="Add new apps")

    def handle(self, *args, **options):
        if not options.get("new_apps"):
            SystemExit("Provide app name...")
        new_apps = options.get("new_apps")
        settings = (
            options["settings"]
            if options.get("settings")
            else os.environ.get("DJANGO_SETTINGS_MODULE")
        )
        scaffold = ScaffoldApp(settings, new_apps)
        scaffold.execute()



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
setup(
    name="manny",
    version="0.1.3",
    author="q0w",
    license="MIT",
    description="Django scaffold like in Ruby on Rails",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/q0w/manny/",
    packages=[
        "scaffold",
        "scaffold.kit",
        "scaffold.management",
        "scaffold.management.commands",
    ],
    python_requires=">=3.6",
    include_package_data=True,
    install_requires=["Django>=1.11", "black", "astor"],
    zip_safe=False,
    keywords="Django scaffold",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django :: 3.0",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP",
    ],
)
